!function() {

"use strict";

const extension  = chrome.runtime.getManifest(),
      noChromeId = -1,
      isChrome = typeof browser == "undefined",
      QS_NONE = 0, // No quest is being loaded.
      QS_LAUNCHED = 1, // Quest has been launched.
      QS_INITIATED = 2, // Content port has connected.
      QS_LOADED = 3,
      BACKUP_PREFIX = "backup_",
      account = {
        avatarName: null,
        username:   null,
        passHash:   null,
        data: {
            fields: [
                "contestIslandMap",
                "contestEmail",
                "Goldfish",
                "HopCandyHunt"
            ],
            default: {
                created: null,
                version: extension.version,
                quests: { }
            },
            firstFieldExtra: 2,  // Extra (non-JSON) data for first field
            maxFieldLength: 499, // Max length of JSON string per field
            base: 36,
            modified: false,
            lastSig: -1, // Signatures prevent malformed JSONs from being created (most of the time)
            // The following null values will be filled in automatically
            maxDataLength: null, // Max JSON length
            value: null
        }
      },
      quest = {
        tab:    noChromeId,
        window: noChromeId,
        name:   null,
        group:  null,
        saves:  null,
        prize:  null,
        state: QS_NONE,
        gamePath: ""
      },
      ports = {
        [ PORT_P ]: null,
        [ PORT_C ]: null,
        [ PORT_E ]: null,
      },
      browserStorage = {
        connected:   false,
        getCallback: null,
        setCallback: null
      },
      validNames = {
        first: [ "", "Angry", "Barefoot", "Bashful", "Beefy", "Bendy", "Big", "Blue", "Bony", "Brave", "Bronze", "Busy", "Calm", "Cheerful", "Chilly", "Clean", "Comical", "Cool", "Crazy", "Creepy", "Cuddly", "Curious", "Dangerous", "Dizzy", "Fast", "Fearless", "Fierce", "Friendly", "Funny", "Gentle", "Giant", "Golden", "Greedy", "Green", "Grey", "Grumpy", "Hamburger", "Happy", "Hungry", "Hyper", "Icy", "Incredible", "Invisible", "Jumpy", "Lazy", "Little", "Lone", "Loud", "Lucky", "Mad", "Magic", "Maroon", "Massive", "Messy", "Mighty", "Moody", "Muddy", "Neat", "Nervous", "Nice", "Noisy", "Orange", "Perfect", "Popular", "Prickly", "Purple", "Quick", "Quiet", "Red", "Rough", "Scary", "Serious", "Shaggy", "Shaky", "Shifty", "Shiny", "Shoeless", "Short", "Shy", "Silly", "Silver", "Skinny", "Sleepy", "Slippery", "Small", "Smart", "Sneaky", "Speedy", "Sporty", "Spotted", "Squeezy", "Sticky", "Strange", "Striped", "Super", "Tall", "Thirsty", "Tough", "Trusty", "White", "Wild", "Yellow", "Young", "Zany", "Zippy" ],
        last:  [ "", "Ant", "Axe", "Bean", "Bear", "Bee", "Beetle", "Belly", "Berry", "Biker", "Bird", "Bite", "Boa", "Bones", "Boot", "Brain", "Bubbles", "Bug", "Burger", "Cactus", "Carrot", "Catfish", "Cheetah", "Chicken", "Claw", "Cloud", "Clown", "Club", "Comet", "Coyote", "Crab", "Crown", "Crumb", "Crush", "Dolphin", "Dragon", "Drummer", "Eagle", "Eel", "Eye", "Fang", "Feather", "Fire", "Fish", "Flame", "Flipper", "Fly", "Flyer", "Foot", "Fox", "Gamer", "Ghost", "Glove", "Goose", "Grape", "Hamburger", "Hammer", "Hawk", "Heart", "Hero", "Hippo", "Hopper", "Horse", "Ice", "Icicle", "Joker", "Jumper", "Kid", "Knuckle", "Leaf", "Leopard", "Lightning", "Lion", "Lizard", "Lobster", "Melon", "Monster", "Moon", "Mosquito", "Noodle", "Octopus", "Onion", "Owl", "Panda", "Paw", "Peanut", "Pear", "Pelican", "Penguin", "Plug", "Popper", "Raptor", "Rider", "Ring", "Rock", "Runner", "Scorpion", "Seagull", "Seal", "Shadow", "Shark", "Shell", "Singer", "Sinker", "Skull", "Skunk", "Sky", "Sneeze", "Snowball", "Socks", "Speck", "Spider", "Spinner", "Sponge", "Star", "Starfish", "Stomper", "Storm", "Sun", "Sword", "Thunder", "Tiger", "Toes", "Tomato", "Tooth", "Tornado", "Tummy", "Turtle", "Typhoon", "Walker", "Whale", "Wing", "Wolf" ]
      }, portNames = {
        [ PORT_P ]: "popup",
        [ PORT_C ]: "content",
        [ PORT_E ]: "external"
      };

let devMode = false,
    config = null,
    launchUrl,
    webListenerRedirects;

const PATH_POP_ORIGIN  = "https://www.poptropica.com",
      PATH_POP = PATH_POP_ORIGIN + "/";

// Set once dev and server config files are retrieved
let PATH_SELF_ORIGIN = "https://quests.idkonpop.com",
    PATH_SELF,
    PATH_SELF_VERSION,
    PATH_SELF_VERSION_JS;

function main() {
    function onDevConfigSet() {
        PATH_SELF = PATH_SELF_ORIGIN + "/";
        resetWebListenerObj();
        account.data.maxDataLength = account.data.maxFieldLength * account.data.fields.length - account.data.firstFieldExtra;
        chrome.runtime.onConnect.addListener(onPortConnect);

        if(isChrome)
            chrome.runtime.onConnectExternal.addListener(port => onPortConnect(port, PORT_E));
    }

    const log = console.log,
          origLaunch = extension.content_scripts[0].matches[0];

    launchUrl = origLaunch.substring(0, origLaunch.length - 1) + "show.php" + QUERY_POP_LAUNCH;
    console.log = (...msg) => devMode && log(...msg);
    request(chrome.runtime.getURL("dev/key"), null, CT_TEXT)
        .then(function(key) {
            PATH_SELF_ORIGIN = "https://quests-dev.idkonpop.com";
            devMode = true;

            function appendHeader(details) {
                details.requestHeaders.push({
                    name: "x-dev-key",
                    value: key
                });
                return { requestHeaders: details.requestHeaders };
            }

            chrome.webRequest.onBeforeSendHeaders.addListener(appendHeader, { urls: [ PATH_SELF_ORIGIN + "/*" ] }, [ "blocking", "requestHeaders" ]);
            onDevConfigSet();
        })
        .catch(onDevConfigSet);
}

function getCurrentTime() {
    const divide = 60000,
          ms = new Date().getTime();

    return (ms - (ms % divide)) / divide;
}

function cloneObject(obj) {
    const ret = { };

    for(let key in obj)
        ret[key] = cloneAtom(obj[key]);

    return ret;
}

function cloneArray(arr) {
    let ret = new arr.constructor(arr.length);

    for(let i = 0; i < arr.length; i++)
        ret[i] = cloneAtom(arr[i]);

    return ret;
}

function cloneAtom(atom) {
    if(typeof atom != "object" || atom === null) // Copy
        return atom;
    else // Reference
        return (Array.isArray(atom) ? cloneArray : cloneObject)(atom);
}

function setUpConfigData(obj) {
    if(obj.maintenance)
        throw ERR_EXT_MNT;
    if(obj.version.client != extension.version)
        throw ERR_EXT_UPD;

    config = obj;
    config.quests.valid = getValidQuestsFromConfigQuests(config.quests);

    if(!config.styles)
        config.styles = { popup: null, game: null };

    PATH_SELF_VERSION = PATH_SELF + (devMode ? "trunk/" : `versions/${ config.version.server }/`);
    PATH_SELF_VERSION_JS = PATH_SELF_VERSION + "modules/";

    if(config.specialMenu)
        config.specialMenu.path = PATH_SELF;
}

function checkForRedeemedCodes() {
    return new Promise(function(onSuccess, onError) {
        const codeArray = [ ],
              questArray = [ ],
              quests = config.quests.active;

        for(let i = 0; i < quests.length; i++) {
            const name = quests[i].code,
                  prize = getQuestData(name).prize;

            if(prize) {
                codeArray.push(prize);
                questArray.push(name);
            }
        }

        if(codeArray.length) {
            apiRequest("checkPrizeRedemptions", codeArray)
                .then(function(redemptionArray) {
                    for(let i = 0; i < redemptionArray.length; i++)
                        if(redemptionArray[i])
                            setQuestData(questArray[i], { prize: "" });

                    saveQuestData()
                        .then(onSuccess)
                        .catch(onError);
                })
                .catch(onError)
        } else
            onSuccess();
    });
}

function getValidQuestsFromConfigQuests(quests) {
    const valid = quests.active.concat(quests.inactive);

    for(let i = 0; i < valid.length; i++) {
        const code = valid[i].code;

        if(valid.indexOf(code) != -1)
            throw new ReferenceError(`Two quests exist with the code "${ code }"`);

        valid[i] = code;
    }

    return valid;
}

function onQuestTerminated() {
    if(quest.name === null)
        return;

    console.log("Quest has been terminated.");
    quest.state = QS_NONE;
    quest.name = quest.group = quest.saves = quest.prize = null;
    quest.gamePath = "";
    quest.tab  = quest.window = noChromeId;
    chrome.tabs.onRemoved.removeListener(onTabRemoved);
    resetWebListenerRedirects();
    onQuestProgressChange(getQuestData().progress);
}

function onQuestProgressChange(status) {
    sendToPort(PORT_P, CMD_POPUP_QUEST_STATUS, quest.name, status, getQuestData(quest.name).prize);
}

function onPortConnect(port, shortName = port.name) {
    if(!config && shortName != PORT_P) {
        port.disconnect();
        return;
    }

    const fullName = portNames[shortName];

    console.log("A %s port has connected.", fullName);
    port.onMessage.addListener(message => console.log("On %s port message: %o", fullName, message));
    port.onDisconnect.addListener(function() {
        ports[shortName] = null;
        console.log("A %s port has disconnected.", fullName);
    });

    switch(shortName) {
        case PORT_P:
            setPopupPort(port);
            break;
        case PORT_C:
            setContentPort(port);
            break;
        case PORT_E:
            setExternalPort(port);
            break;
        default:
            throw new Error(`Invalid port type "${ shortName }"`);
    }
}

function sendToPort(id, ...data) {
    if(id != PORT_E || data[0] || data[1] != "webLoad")
        console.log("Sending to %s port:", portNames[id], data);

    if(ports[id])
        ports[id].postMessage(data);
}

function disconnectPort(id) {
    console.log("Disconnecting a %s port.", portNames[id]);

    if(ports[id]) {
        ports[id].disconnect();
        ports[id] = null;
    }
}

function setPopupPort(port) {
    if(ports[PORT_P])
        ports[PORT_P].disconnect();

    ports[PORT_P] = port;
    port.onMessage.addListener(onPopupPortMessage);

    const onError = err => sendToPort(PORT_P, null, err),
          setupComplete = () => checkForRedeemedCodes()
            .then(() => sendToPort(PORT_P, null,
                null,
                isChrome,
                quest.state != QS_NONE ? quest.name : null,
                config.quests.useScroll,
                config.styles.popup || null,
                (account.username === null) ? ARG_POPUP_INIT_SECTION_LOGIN : ARG_POPUP_INIT_SECTION_MAIN,
                config.specialMenu
            ))
            .catch(onError);

    request(PATH_SELF + "config.json", null, CT_JSON)
        .then(function(obj) {
            setUpConfigData(obj);

            if(browserStorage.send)
                setupComplete();
            else
                loadDiskGateway()
                    .then(setupComplete)
                    .catch(onError);
        })
        .catch(onError);
}

function setContentPort(port) {
    const toSend = [ null, devMode, isChrome, quest.gamePath, PATH_SELF_VERSION, PATH_SELF_VERSION_JS, "noQuest", config.styles.game || null ];

    if(ports[PORT_C])
        disconnectPort(PORT_C);

    ports[PORT_C] = port;
    port.onMessage.addListener(onContentPortMessage);

    if((quest.name !== null || typeof quest.saves == "boolean" && !quest.saves) && quest.state != QS_LOADED) {
        quest.state = QS_INITIATED;
        onQuestProgressChange(PROGRESS_PLAYING);
        toSend[toSend.length - 2] = quest.group;
    }

    sendToPort(PORT_C, ...toSend);
}

function setExternalPort(port) {
    if(!ports[PORT_E] && quest.name !== null && quest.state == QS_INITIATED) {
        quest.state = QS_LOADED;
        disconnectPort(PORT_C);
        ports[PORT_E] = port;
        port.onMessage.addListener(onExternalPortMessage);
        port.onDisconnect.addListener(onQuestTerminated);
        sendToPort(PORT_E, false, null,
            account.username,
            account.passHash,
            quest.name,
            quest.prize,
            getQuestData()
        );
    } else
        port.disconnect();
}

function onPopupPortMessage(message) {
    const [ cmd, ...data ] = message,
          sendResponse = (...args) => sendToPort(PORT_P, cmd, ...args);

    switch(cmd) {
        case CMD_POPUP_LOGOUT:
            logOut();
            sendResponse();
            break;
        case CMD_POPUP_LOGIN:
            logIn(...data)
                .then(() => sendResponse(null))
                .catch(sendResponse);
            break;
        case CMD_POPUP_GET_QUESTS:
            sendResponse(account.avatarName, getQuestArrayForPopup());
            break;
        case CMD_POPUP_QUEST_LAUNCH:
            launchQuest(data[0])
                .then(() => sendResponse(null))
                .catch(sendResponse);
            break;
        case CMD_POPUP_QUEST_TAB:
            chrome.tabs.update(quest.tab, { active: true }, () => chrome.windows.update(quest.window, { focused: true }, sendResponse));
            break;
        default:
            throw new Error(`Unknown popup command ${ cmd }`);
    }
}

function onContentPortMessage(message) {
    const [ cmd ] = message;

    switch(cmd) {
        case CMD_CONTENT_SWITCH_TO_EXT:
            const port = ports[PORT_C];
            port.onMessage.removeListener(onContentPortMessage);
            ports[PORT_C] = null;
            onPortConnect(port, PORT_E);
            break;
    }
}

function onExternalPortMessage(message) {
    if(quest.name === null)
        return;

    const [ , id, cmd, ...data ] = message,
          sendResponse = (error = null, ...info) => sendToPort(PORT_E, true, id, error, ...info);

    switch(cmd) {
        case "setQuestData":
            if(!quest.saves)
                return sendResponse();

            setQuestData(quest.name, {
                progress: PROGRESS_STARTED,
                data: data[0]
            });
            saveQuestData()
                .then(sendResponse)
                .catch(sendResponse);
            break;
        case "wonQuest":
            completeQuest()
                .then(data => sendResponse(null, data))
                .catch(sendResponse);
            break;
        case "addRedirect":
            addWebListenerRedirect(...data);
            sendResponse();
            break;
        case "removeRedirect":
            removeWebListenerRedirect(data[0]);
            sendResponse();
            break;
        case "apiRequest":
            apiRequest(data[0], data[1])
                .then(data => sendResponse(null, data))
                .catch(sendResponse);
            break;
        default:
            throw new Error(`Unknown external command "${ cmd }".`);
    }
}

function onTabRemoved(tabId) {
    if(tabId == quest.tab)
        onQuestTerminated();
}

function addWebListenerCallback() {
    chrome.webRequest.onBeforeRequest.addListener(onWebListenerRequest, {
        urls: [ extension.permissions[extension.permissions.length - 1] ],
        tabId: quest.tab
    }, [ "blocking" ]);

    console.log("Ready for web listeners.");
}

function resetWebListenerRedirects() {
    chrome.webRequest.onBeforeRequest.removeListener(onWebListenerRequest);
    resetWebListenerObj();
    console.log("Shut down web listening.");
}

function resetWebListenerObj() {
    webListenerRedirects = { };
}

function addWebListenerRedirect(url, url2) {
    if(webListenerRedirects.hasOwnProperty(url))
        throw new Error(`Multiple redirects cannot be set for URL "${ url }" (already going to "${ webListenerRedirects[url] }", tried setting to "${ url2 }").`);

    webListenerRedirects[url] = url2;
}

function removeWebListenerRedirect(url) {
    if(!webListenerRedirects.hasOwnProperty(url))
        throw new Error(`No redirect for URL "${ url }" exists.`);

    delete webListenerRedirects[url];
}

function onWebListenerRequest(request) {
    const obj = { };

    if(request.url.startsWith(quest.gamePath)) {
        const url = request.url.substring(quest.gamePath.length),
              redirUrl = webListenerRedirects[url];

        if(typeof redirUrl != "undefined") {
            if(redirUrl === null)
                obj.cancel = true;
            else
                obj.redirectUrl = redirUrl;

            sendToPort(PORT_E, false, "webRedirect", url);
        }

        sendToPort(PORT_E, false, "webLoad", url);
    } else {
        const trackerPrefix = PATH_POP + "brain/track.php?";

        if(request.url.startsWith(trackerPrefix))
            sendToPort(PORT_E, false, "webTrack", request.url.substring(trackerPrefix.length))
    }

    return obj;
}

function launchQuest(name) {
    return new Promise(function(onSuccess_unused, onError) {
        const qConfig = getQuestConfig(name),
              loadQuest = () => chrome.tabs.create({ url: launchUrl }, function(tab) {
                // These values must be set after the previous quest tab (if there is one) has been closed.
                quest.tab    = tab.id;
                quest.window = tab.windowId;
                quest.group  = qConfig ? qConfig.group : name;
                quest.name   = name;
                quest.saves  = qConfig ? qConfig.saves : false;
                quest.prize  = qConfig && qConfig.prize ? qConfig.prize.type : null;
                quest.state  = QS_LAUNCHED;
                chrome.tabs.onRemoved.addListener(onTabRemoved);
                addWebListenerCallback();
                ports[PORT_P] = null;
              });

        function resetCheck() {
            if(getQuestData(name).progress == PROGRESS_COMPLETE)
                setQuestData(name, { progress: PROGRESS_NONE });

            setGamePath();
        }

        function setGamePath() {
            safeFetch(PATH_POP + "haxe/", null)
                .then(function(resp) {
                    if(!resp.url.startsWith(PATH_POP))
                        return onError(ERR_REQ_SERVER);

                    quest.gamePath = resp.url;
                    loadQuest();
                })
                .catch(onError);
        }

        if(quest.tab == noChromeId)
            resetCheck();
        else {
            // Sometimes Chrome says the quest tab has been removed before all of its ports have disconnected, so we need to reset the quest session data before removing the tab.
            const tab = quest.tab; // Copy the quest tab ID before the quest session data is reset.

            onQuestTerminated();
            chrome.tabs.remove(tab, resetCheck);
        }
    });
}

function getQuestConfig(name) {
    const quests = config.quests.active;

    for(let i = 0; i < quests.length; i++)
        if(quests[i].code === name)
            return quests[i];

    return null;
}

// Move this function to within `request`?
function getContentType(data) {
    const type = typeof data;

    switch(type) {
        case "string":
            return CT_QUERY;
        case "object":
            return CT_JSON;
        default:
            throw new Error(`No content type exists for primitive type "${ type }"`);
    }
}

function safeFetch(url, opts) {
    return new Promise(function(onSuccess, onError) {
        if(typeof opts != "object" || opts === null)
            opts = { };

        opts.cache = "no-store"; // Avoid caching! A status code in the 300 range will cause `Response.ok` to be false.

        fetch(url, opts)
            .then(function(response) {
                if(!response.ok)
                    throw ERR_REQ_SERVER;
                else
                    onSuccess(response);
            }).catch(function onCatch(err) {
                // The `fetch` function throws an error in the case of a network error. We want to know when a network error occurs, but it must be ERR_REQ_NETWORK.
                // It's also important to keep in mind that other, valid errors may be thrown as well.
                if(typeof err == "number")
                    onError(err);
                else
                    onError(ERR_REQ_NETWORK);
            });
    });
}

function request(url, data, respCT = null) {
    return new Promise(function(onSuccess, onError) {
        const isGet = data === null,
              cType = isGet ? null : getContentType(data);

        safeFetch(url, isGet ? null : {
            method: "POST",
            headers: { "Content-Type": cType },
            credentials: "omit",
            body: (cType == CT_JSON ? JSON.stringify(data) : data)
        }).then(function onResponse(response) {
            switch(respCT) {
                case CT_TEXT:
                    response.text()
                        .then(str => onSuccess(str)); // `Body.text` should never throw an error for successful requests.
                    break;
                case CT_JSON:
                    response.json()
                        .then(onSuccess)
                        .catch(() => onError(ERR_REQ_RESP_ENCODING));
                    break;
                case CT_QUERY:
                    response.text()
                        .then(str => onSuccess(new URLSearchParams(str))); // `Body.text` should never throw an error for successful requests.
                    break;
                default:
                    if(!respCT)
                        onSuccess();
                    else
                        console.error(`Unknown response type of "${ respCT }"`);
            }
        }).catch(function onCatch(err) {
            // The `fetch` function throws an error in the case of a network error. We want to know when a network error occurs, but it must be ERR_REQ_NETWORK.
            // It's also important to keep in mind that other, valid errors may be thrown as well.
            if(typeof err != "number")
                err = ERR_REQ_NETWORK;

            onError(err);
        });
    });
}

function apiRequest(method, data) {
    return new Promise(function(onSuccess, onError) {
        function onResponse(obj) {
            if(obj.success)
                onSuccess(obj.data);
            else switch(obj.data) {
                case "badVersion":
                    onError(ERR_EXT_UPD);
                    break;
                case "maintenance":
                    onError(ERR_EXT_MNT);
                    break;
                case "tryAgain":
                    tryRequest();
                    break;
                default:
                    onError(ERR_REQ_SERVER);
                    break;
            }
        }

        let tries = -1;

        function tryRequest() {
            if(++tries == 3)
                onError(ERR_REQ_SERVER);
            else
                request(PATH_SELF + "api/", {
                    method: method,
                    data: data,
                    version: config.version.client
                }, CT_JSON)
                    .then(onResponse)
                    .catch(onError);
        }

        tryRequest();
    });
}

function logIn(user, password) {
    return new Promise(function(onSuccess, onError) {
        user = user.toLowerCase();

        const passHash = md5(password.toLowerCase());

        getUserData(user, passHash)
            .then(function(data) {
                function setData(value, sig) {
                    account.data.value = value;
                    account.data.lastSig = sig;
                }

                account.avatarName = validateAvatarName(data.firstName, data.lastName);
                account.username   = user;
                account.passHash   = passHash;
                account.data.default.created = getCurrentTime();

                const [ cloudValue, cloudSig, cloudNeedUpdate ] = unpackQuestData(data, true);

                if(cloudValue) {
                    setData(cloudValue, cloudSig);
                    updateQuestData(cloudNeedUpdate)
                        .then(onSuccess)
                        .catch(onError);
                } else
                    getQuestDataFromDisk()
                        .then(function(localData) {
                            const [ localValue, localSig, localNeedUpdate ] = unpackQuestData(localData, false);
                            setData(localValue, localSig);
                            console.log("Retrieved backup data (no valid cloud data)");

                            updateQuestData(localNeedUpdate)
                                .then(onSuccess)
                                .catch(onError);
                        })
                        .catch(onError);
            })
            .catch(onError);
    });
}

function logOut() {
    account.avatarName =
    account.username   =
    account.passHash   =
    account.data.value =
    account.data.default.created = null;

    account.data.modified = false;
    account.data.lastSig  = -1;

    onQuestTerminated();
    disconnectPort(PORT_C);
    disconnectPort(PORT_E);
}

function validateAvatarName(fName, lName) {
    return (validNames.first.indexOf(fName) == -1 ||
            validNames.last.indexOf(lName)  == -1 ||
           (fName === "" && lName === "")) ? "Unnamed User" : `${ fName } ${ lName }`;
}

function getQuestArrayForPopup() {
    const quests = cloneArray(config.quests.active);

    for(let i = 0; i < quests.length; i++) {
        const current = quests[i];

        if(current.prize)
            current.prize.value = getQuestData(current.code).prize;

        delete current.group; // Remove unnecessary data to speed up serialization
        current.progress = getQuestData(current.code).progress;
    }

    return quests;
}

function getUserData(user = account.username, passHash = account.passHash) {
    return new Promise(function(onSuccess, onError) {
        const params = new URLSearchParams();

        params.set("login", user);
        params.set("pass_hash", passHash);

        request(`${ PATH_POP }login.php`, params.toString(), CT_QUERY)
            .then(function onResponse(body) {
                const P_ANS  = "answer",
                      P_JSON = "json";

                if(body.get(P_ANS) != "ok" || !body.has(P_JSON))
                    throw ERR_POP_AUTH;

                let json;

                try {
                    json = JSON.parse(body.get(P_JSON));
                } catch(e) {
                    throw ERR_POP_AUTH;
                }

                const data = (json.userData == "[]") ? { } : JSON.parse(json.userData);

                data.firstName = json.firstname;
                data.lastName  = json.lastname;

                onSuccess(data);
            })
            .catch(onError);
    });
}

function encodeJSON(str) {
    str = str.replace(/\{/g, "(")
             .replace(/\}/g, ")")
             .replace(/\\/g, "/")
             .replace(/"/g, "'");

    return str;
}

function decodeJSON(str) {
    str = str.replace(/\(/g, "{")
             .replace(/\)/g, "}")
             .replace(/\//g, "\\")
             .replace(/'/g, '"');

    return str;
}

async function updateQuestData(needUpdate) {
    if(needUpdate) {
        account.data.modified = true;
        await saveQuestData();
    }
}

function unpackQuestData(allData, fromCloud) {
    const defaultReturn = [ // Return this at the first sight of trouble!
            fromCloud ? null : account.data.default,
            -1,
            false
          ];

    let jsonString,
        sig,
        needUpdate;

    if(fromCloud) {
        const fields = account.data.fields,
              firstValue = allData[fields[0]],
              firstFieldExtra = account.data.firstFieldExtra,
              base = account.data.base;

        if(typeof firstValue != "string" || firstValue.length <= 2)
            return defaultReturn;

        const numFields = parseInt(firstValue.substring(0, firstFieldExtra), base) + 1;

        sig = parseInt(firstValue[firstFieldExtra], base) - 1;

        if(Number.isNaN(sig))
            return defaultReturn;

        jsonString = firstValue.substring(firstFieldExtra + 1);

        for(let i = 1; i < numFields; i++) {
            const value = allData[fields[i]];

            if(typeof value != "string" || parseInt(value[0], base) != sig)
                return defaultReturn;
            else
                jsonString += value.substring(1);
        }
    } else {
        if(typeof allData != "string" || allData.length <= 1)
            return defaultReturn;

        sig = parseInt(allData[0], account.data.base) - 1;

        if(Number.isNaN(sig))
            return defaultReturn;

        jsonString = allData.substring(1);
    }

    let json = null;

    try {
        json = JSON.parse(decodeJSON(jsonString));
    } catch(err) { }

    if(typeof json != "object" || json == null || Array.isArray(json) ||
       typeof json.v != "string" ||
       typeof json.c != "number" ||
       typeof json.q  != "object" || json.q == null || Array.isArray(json.q))
        return defaultReturn;

    json.version = json.v;
    json.created = json.c;
    json.quests = json.q;
    delete json.v;
    delete json.c;
    delete json.q;

    for(let key in json.quests) {
        const quest = json.quests[key];

        if(typeof quest.s == "undefined")
            quest.progress = PROGRESS_COMPLETE;
        else {
            quest.progress = quest.s;
            delete quest.s;
        }

        if(typeof quest.c == "undefined")
            quest.completions = 0;
        else {
            quest.completions = quest.c;
            delete quest.c;
        }

        if(typeof quest.d == "undefined")
            quest.data = null;
        else {
            quest.data = quest.d;
            delete quest.d;
        }

        if(typeof quest.p == "undefined")
            quest.prize = null;
        else {
            quest.prize = quest.p;
            delete quest.p;
        }
    }

    needUpdate = json.version != extension.version;
    json.version = extension.version;

    removeDeadQuestDataFromQuestObj(json.quests);

    return [ json, sig, needUpdate ];
}

function packQuestData(value, lastSig) {
    value = cloneObject(value);

    for(let key in value.quests) {
        const quest = value.quests[key];

        if(quest.progress != PROGRESS_COMPLETE)
            quest.s = quest.progress;

        if(quest.completions != 0)
            quest.c = quest.completions;

        if(quest.data !== null)
            quest.d = quest.data;

        if(quest.prize !== null)
            quest.p = quest.prize;

        delete quest.progress;
        delete quest.completions;
        delete quest.data;
        delete quest.prize;
    }

    value.q = value.quests;
    value.c = value.created;
    value.v = value.version;
    delete value.quests;
    delete value.created;
    delete value.version;

    const fieldValues    = [ ],
          maxValueLength = account.data.maxFieldLength,
          base   = account.data.base,
          newSig = (lastSig + 1) % base,
          newSigString = newSig.toString(base),
          jsonString   = encodeJSON(JSON.stringify(value)),
          firstFieldExtra = account.data.firstFieldExtra;

    if(jsonString.length > account.data.maxDataLength)
        throw ERR_SAVE_MAX;

    fieldValues.push(newSigString + jsonString.substring(0, maxValueLength - firstFieldExtra));

    for(let i = maxValueLength - firstFieldExtra; i < jsonString.length; ) {
        const newCut = jsonString.substring(i, i + maxValueLength);

        fieldValues.push(newSigString + newCut);
        i += newCut.length;
    }

    fieldValues[0] = (fieldValues.length - 1).toString(account.data.base).padStart(firstFieldExtra, "0") + fieldValues[0];

    console.log("Packing save data across %i field(s). Last signature was %i, so this will be %i. Total JSON data used: %i/%i bytes", fieldValues.length, lastSig, newSig, jsonString.length, account.data.maxDataLength);

    return [
        account.data.fields.slice(0, fieldValues.length),
        fieldValues,
        newSig
    ];
}

function removeDeadQuestDataFromQuestObj(data) {
    const dataQuests  = Object.keys(data),
          validQuests = config.quests.valid;

    for(let i = 0; i < dataQuests.length; i++) {
        const config = getQuestConfig(dataQuests[i]);

        if(validQuests.indexOf(dataQuests[i]) == -1 || !config || !config.saves) {
            delete data[dataQuests[i]];
            console.log('Deleting dead quest "%s"', dataQuests[i]);
        }
    }
}

function completeQuest(name = quest.name) {
    return new Promise(function(onSuccess, onError) {
        const questData = getQuestData(name),
              questPrize = getQuestConfig(name).prize;
        let portData = null,
            redeem = null;

        function finalStep() {
            ++questData.completions;
            setQuestData(name, {
                progress: PROGRESS_COMPLETE,
                data: null,
                prize: redeem
            });
            onQuestProgressChange(PROGRESS_COMPLETE);
            saveQuestData()
                .then(() => onSuccess(portData))
                .catch(onError);
        }

        switch(questPrize.type) {
            case PRIZE_NONE:
                Promise.resolve()
                    .then(finalStep)
                    .catch(onError);
                break;
            case PRIZE_CREDITS:
                portData = questPrize.data;
                addCredits(portData)
                    .then(finalStep)
                    .catch(onError);
                break;
            case PRIZE_ROLE:
                addServerPrize(name)
                    .then(function(code) {
                        redeem = code;
                        portData = Object.assign({ code: code }, questPrize.data);
                        finalStep();
                    })
                    .catch(onError);
                break;
            case PRIZE_ROLE_CODELESS:
                addCodelessServerPrize(name)
                    .then(finalStep)
                    .catch(onError);
                break;
            default:
                throw new Error(`Unknown prize type "${ questPrize.type }".`);
        }
    });
}

function addServerPrize(name) {
    const data = getQuestData(name);
    return data.prize !== null ? Promise.resolve(data.prize) : apiRequest("winPrize", name);
}

function addCodelessServerPrize(name) {
    return apiRequest("winCodelessPrize", {
        user: account.username,
        quest: name
    });
}

function redeemServerPrize(code) {
    return apiRequest("redeemPrize", {
        code: code,
        key: null
    });
}

function addCredits(amount) {
    const db = [
        {
            amount: 250,
            isWheel: true,
            data: 0
        }, {
            amount: 175,
            isWheel: true,
            data: 5
        }, {
            amount: 150,
            isWheel: true,
            data: 4
        }, {
            amount: 100,
            isWheel: true,
            data: 3
        }, {
            amount: 60,
            isWheel: false,
            data: "hard_1"
        }, {
            amount: 50,
            isWheel: true,
            data: 2
        }, {
            amount: 45,
            isWheel: false,
            data: "hard_2"
        }, {
            amount: 40,
            isWheel: false,
            data: "normal_1"
        }, {
            amount: 30,
            isWheel: false,
            data: "hard_3"
        }, {
            amount: 25,
            isWheel: true,
            data: 1
        }, {
            amount: 20,
            isWheel: false,
            data: "normal_3"
        }, {
            amount: 15,
            isWheel: false,
            data: "hard_4"
        }, {
            amount: 10,
            isWheel: false,
            data: "normal_4"
        }, {
            amount: 5,
            isWheel: false,
            data: "easy_4"
        }
    ], transactions = [ ];

    if(amount % 5 != 0 || amount <= 0)
        throw new RangeError("Amount of credits must be a positive integer divisible by five.");

    let i = 0;

    while(amount)
        if(db[i].amount <= amount) {
            transactions.push(db[i]);
            amount -= db[i].amount;
        } else
            i++;

    return new Promise(function(onSuccess, onError) {
        i = -1;

        function next() {
            if(++i == transactions.length)
                onSuccess();
            else {
                const current = transactions[i],
                      onThen = credits => credits == current.amount ? next() : onError(ERR_REQ_RESP_ENCODING);

                if(current.isWheel) {
                    const array = new Array(current.data).fill(null);

                    array.push({
                        type: "credits",
                        id: current.data ? "credits" + current.data : "grandPrize",
                        prize: current.amount
                    });

                    sendToCallRouter("getRandomWinner", array)
                        .then(onThen)
                        .catch(onError);
                } else
                    completeArcadeGame("reality2_" + current.data, current.amount)
                        .then(onThen)
                        .catch(onError);
            }
        }

        next();
    });
}

function completeArcadeGame(name, score) {
    const params = new URLSearchParams();

    params.set("gamename", name);
    params.set("game_time", 0);
    params.set("category", "weekly_highscore");
    params.set("username", account.username);
    params.set("passhash", account.passHash);
    params.set("dbid", 1);
    params.set("score", score);
    params.set("win", 1);
    params.set("loss", 0);
    params.set("personal_high_score", 100);

    return new Promise(function(onSuccess, onError) {
        request(`${ PATH_POP }games/interface/submit_game.php`, params.toString(), CT_QUERY)
            .then(function(resp) {
                const answer = resp.get("answer");

                if(answer == "ok")
                    onSuccess(parseInt(resp.get("credits")));
                else {
                    if(answer == "no-such-user")
                        onError(ERR_POP_AUTH);
                    else
                        onError(ERR_REQ_RESP_ENCODING);
                }
            })
            .catch(onError);
    });
}

function sendToCallRouter(method, arg) {
    return new Promise(function(onSuccess, onError) {
        request(`${ PATH_POP }AMFPHP/callRouter.php`, {
            serviceName: "PlayerService",
            methodName: "fetch",
            parameters: [ method, {
                login: account.username,
                pass_hash: account.passHash,
                dbid: 1
            }, arg ]
        }, CT_JSON)
            .then(obj => obj._explicitType != "PopError" ? onSuccess(obj.value) : onError(ERR_POP_AUTH))
            .catch(onError);
    });
}

function getQuestData(name = quest.name) {
    const defaultReturn = {
        completions: 0,
        progress: PROGRESS_NONE,
        data: null,
        prize: null
    };

    if(account.username === null)
        return defaultReturn;

    const quests = account.data.value.quests;

    if(!quests.hasOwnProperty(name))
        return defaultReturn;

    return quests[name];
}

function setQuestData(name, questData) {
    const questObj = account.data.value.quests;

    questObj[name] = Object.assign(getQuestData(name), questData);
    account.data.modified = true;
}

function saveQuestData(questData = account.data.value, lastSig = account.data.lastSig) {
    return new Promise(function(onSuccess, onError) {
        if(!account.data.modified)
            return onSuccess();

        const [ names, values, newSig ] = packQuestData(questData, lastSig),
              saveToDiskAndFinish = () => saveQuestDataToDisk(questData, newSig)
                .then(function() {
                    account.data.modified = false;
                    console.log("Data save complete!");
                    onSuccess();
                })
                .catch(onError);

        console.log("Saving data...");
        account.data.lastSig = newSig;

        saveQuestDataToCloud(names, values)
            .then(saveToDiskAndFinish)
            .catch(onError);
    });
}

function saveQuestDataToCloud(names, values) {
    return new Promise(function(onSuccess, onError) {
        let i = 0;

        function nextReq() {
            saveFieldToCloud(names[i], values[i])
                .then(() => (++i == values.length) ? onSuccess() : nextReq())
                .catch(onError);
        }

        nextReq();
    });
}

function saveFieldToCloud(name, value) {
    return new Promise(function(onSuccess, onError) {
        const params = new URLSearchParams();

        params.set("field", name);
        params.set("value", value);
        params.set("login", account.username);
        params.set("pass_hash", account.passHash);
        params.set("dbid", "1");

        request(`${ PATH_POP }set_user_field.php`, params.toString(), CT_QUERY)
            .then(resp => resp.get("error") === "" ? onSuccess() : onError(ERR_POP_AUTH))
            .catch(onError);
    });
}

function loadDiskGateway() {
    return new Promise(function(onSuccess, onError) {
        let initiator = chrome.runtime.getURL("");

        initiator = initiator.substring(0, initiator.length - 1);

        function onRequestError(req) {
            if(req.initiator != initiator)
                return;

            removeListeners();
            onError(ERR_REQ_NETWORK);
        }

        function onRequestComplete(req) {
            if(req.initiator != initiator)
                return;

            removeListeners();

            if(req.statusCode != 200 && !req.fromCache)
                onError(ERR_REQ_SERVER);

            // Let's wait for a message back from the gateway before onSuccess is called.
        }

        function removeListeners() {
            chrome.webRequest.onCompleted.removeListener(onRequestComplete);
            chrome.webRequest.onErrorOccurred.removeListener(onRequestError);
        }

        const gatewayUrl = `${ PATH_SELF }storage.html`,
              filterObj  = {
                  urls: [ gatewayUrl ],
                  windowId: noChromeId
              },
              iframe = document.createElement("iframe");

        window.onmessage = function(evt) {
            if(evt.origin != PATH_SELF_ORIGIN)
                throw new Error("Invalid message origin.");
            else if(browserStorage.send) {
                const [ isSetting, data ] = evt.data;

                isSetting ? browserStorage.setCallback() : browserStorage.getCallback(data);
            } else {
                browserStorage.send = (...data) => iframe.contentWindow.postMessage(data, PATH_SELF_ORIGIN);
                console.log("Loaded browser storage gateway.");
                onSuccess();
            }
        };

        iframe.src = gatewayUrl;
        document.body.appendChild(iframe);
        chrome.webRequest.onCompleted.addListener(onRequestComplete, filterObj);
        chrome.webRequest.onErrorOccurred.addListener(onRequestError, filterObj);
    });
}

function getQuestDataFromDisk() {
    return new Promise(function(onSuccess) {
        if(browserStorage.getCallback)
            throw new Error("Cannot request backup data for more than one user at a time.");

        browserStorage.getCallback = function(data) {
            browserStorage.getCallback = null;
            onSuccess(data);
        };
        browserStorage.send(false, BACKUP_PREFIX + account.username);
    });
}

function saveQuestDataToDisk(questData, newSig) {
    return new Promise(function(onSuccess) {
        if(browserStorage.setCallback)
            throw new Error("Cannot request setting backup data for more than one user at a time.");

        browserStorage.setCallback = function() {
            browserStorage.setCallback = null;
            onSuccess();
        };
        browserStorage.send(true, BACKUP_PREFIX + account.username, newSig.toString(account.data.base) + encodeJSON(JSON.stringify(questData)));
    });
}

// UNSAFE - development only, exposed to debug console for emergencies
window.wipeDataAndRestart = function(localOnly = false) {
    if(!config || account.username === null)
        return "No one is logged in.";

    const totalQuests = config.quests.active,
          toRedeem = [ ];

    for(let i = 0; i < totalQuests.length; i++) {
        const data = getQuestData(totalQuests[i].code);

        if(data.prize)
            toRedeem.push(data.prize);
    }

    let i = -1;
    function next() {
        if(++i == toRedeem.length)
            saveQuestDataToDisk(null, "")
                .then(function() {
                    const endFunc = () => chrome.runtime.reload();

                    if(localOnly)
                        endFunc();
                    else
                        saveFieldToCloud(account.data.fields[0], "")
                            .then(endFunc);
                });
        else
            redeemServerPrize(toRedeem[i])
                .then(next)
                .catch(next);
    }

    const delaySeconds = 10;

    setTimeout(next, delaySeconds * 1000);
    return `The data reset will commence in ${ delaySeconds } seconds. Custom Quests will reload when it finishes.`;
};

// Function name is `md5`
!function(n){"use strict";function t(n,t){var r=(65535&n)+(65535&t);return(n>>16)+(t>>16)+(r>>16)<<16|65535&r}function r(n,t){return n<<t|n>>>32-t}function e(n,e,o,u,c,f){return t(r(t(t(e,n),t(u,f)),c),o)}function o(n,t,r,o,u,c,f){return e(t&r|~t&o,n,t,u,c,f)}function u(n,t,r,o,u,c,f){return e(t&o|r&~o,n,t,u,c,f)}function c(n,t,r,o,u,c,f){return e(t^r^o,n,t,u,c,f)}function f(n,t,r,o,u,c,f){return e(r^(t|~o),n,t,u,c,f)}function i(n,r){n[r>>5]|=128<<r%32,n[14+(r+64>>>9<<4)]=r;var e,i,a,d,h,l=1732584193,g=-271733879,v=-1732584194,m=271733878;for(e=0;e<n.length;e+=16)i=l,a=g,d=v,h=m,g=f(g=f(g=f(g=f(g=c(g=c(g=c(g=c(g=u(g=u(g=u(g=u(g=o(g=o(g=o(g=o(g,v=o(v,m=o(m,l=o(l,g,v,m,n[e],7,-680876936),g,v,n[e+1],12,-389564586),l,g,n[e+2],17,606105819),m,l,n[e+3],22,-1044525330),v=o(v,m=o(m,l=o(l,g,v,m,n[e+4],7,-176418897),g,v,n[e+5],12,1200080426),l,g,n[e+6],17,-1473231341),m,l,n[e+7],22,-45705983),v=o(v,m=o(m,l=o(l,g,v,m,n[e+8],7,1770035416),g,v,n[e+9],12,-1958414417),l,g,n[e+10],17,-42063),m,l,n[e+11],22,-1990404162),v=o(v,m=o(m,l=o(l,g,v,m,n[e+12],7,1804603682),g,v,n[e+13],12,-40341101),l,g,n[e+14],17,-1502002290),m,l,n[e+15],22,1236535329),v=u(v,m=u(m,l=u(l,g,v,m,n[e+1],5,-165796510),g,v,n[e+6],9,-1069501632),l,g,n[e+11],14,643717713),m,l,n[e],20,-373897302),v=u(v,m=u(m,l=u(l,g,v,m,n[e+5],5,-701558691),g,v,n[e+10],9,38016083),l,g,n[e+15],14,-660478335),m,l,n[e+4],20,-405537848),v=u(v,m=u(m,l=u(l,g,v,m,n[e+9],5,568446438),g,v,n[e+14],9,-1019803690),l,g,n[e+3],14,-187363961),m,l,n[e+8],20,1163531501),v=u(v,m=u(m,l=u(l,g,v,m,n[e+13],5,-1444681467),g,v,n[e+2],9,-51403784),l,g,n[e+7],14,1735328473),m,l,n[e+12],20,-1926607734),v=c(v,m=c(m,l=c(l,g,v,m,n[e+5],4,-378558),g,v,n[e+8],11,-2022574463),l,g,n[e+11],16,1839030562),m,l,n[e+14],23,-35309556),v=c(v,m=c(m,l=c(l,g,v,m,n[e+1],4,-1530992060),g,v,n[e+4],11,1272893353),l,g,n[e+7],16,-155497632),m,l,n[e+10],23,-1094730640),v=c(v,m=c(m,l=c(l,g,v,m,n[e+13],4,681279174),g,v,n[e],11,-358537222),l,g,n[e+3],16,-722521979),m,l,n[e+6],23,76029189),v=c(v,m=c(m,l=c(l,g,v,m,n[e+9],4,-640364487),g,v,n[e+12],11,-421815835),l,g,n[e+15],16,530742520),m,l,n[e+2],23,-995338651),v=f(v,m=f(m,l=f(l,g,v,m,n[e],6,-198630844),g,v,n[e+7],10,1126891415),l,g,n[e+14],15,-1416354905),m,l,n[e+5],21,-57434055),v=f(v,m=f(m,l=f(l,g,v,m,n[e+12],6,1700485571),g,v,n[e+3],10,-1894986606),l,g,n[e+10],15,-1051523),m,l,n[e+1],21,-2054922799),v=f(v,m=f(m,l=f(l,g,v,m,n[e+8],6,1873313359),g,v,n[e+15],10,-30611744),l,g,n[e+6],15,-1560198380),m,l,n[e+13],21,1309151649),v=f(v,m=f(m,l=f(l,g,v,m,n[e+4],6,-145523070),g,v,n[e+11],10,-1120210379),l,g,n[e+2],15,718787259),m,l,n[e+9],21,-343485551),l=t(l,i),g=t(g,a),v=t(v,d),m=t(m,h);return[l,g,v,m]}function a(n){var t,r="",e=32*n.length;for(t=0;t<e;t+=8)r+=String.fromCharCode(n[t>>5]>>>t%32&255);return r}function d(n){var t,r=[];for(r[(n.length>>2)-1]=void 0,t=0;t<r.length;t+=1)r[t]=0;var e=8*n.length;for(t=0;t<e;t+=8)r[t>>5]|=(255&n.charCodeAt(t/8))<<t%32;return r}function h(n){return a(i(d(n),8*n.length))}function l(n,t){var r,e,o=d(n),u=[],c=[];for(u[15]=c[15]=void 0,o.length>16&&(o=i(o,8*n.length)),r=0;r<16;r+=1)u[r]=909522486^o[r],c[r]=1549556828^o[r];return e=i(u.concat(d(t)),512+8*t.length),a(i(c.concat(e),640))}function g(n){var t,r,e="";for(r=0;r<n.length;r+=1)t=n.charCodeAt(r),e+="0123456789abcdef".charAt(t>>>4&15)+"0123456789abcdef".charAt(15&t);return e}function v(n){return unescape(encodeURIComponent(n))}function m(n){return h(v(n))}function p(n){return g(m(n))}function s(n,t){return l(v(n),v(t))}function C(n,t){return g(s(n,t))}function A(n,t,r){return t?r?s(t,n):C(t,n):r?m(n):p(n)}"function"==typeof define&&define.amd?define(function(){return A}):"object"==typeof module&&module.exports?module.exports=A:n.md5=A}(window);

main();

}();