"use strict";

const title = "Loading...";
let connected = false;

function main() {
    if(location.search == QUERY_POP_LAUNCH && !isQuestInitiated()) {
        document.title = title;
        createPort(PORT_C, disconnect, onStartup);
    }
}

function isQuestInitiated() {
    return document.title === "Custom Quests";
}

function createEventPort() {
    const evtName = "message",
          STATE_UNCONNECTED = 0,
          STATE_CONNECTED = 1,
          STATE_DISCONNECTED = 2;

    let state = STATE_UNCONNECTED;

    function onExternalMessage(obj) {
        const { data } = obj;

        if(!Array.isArray(data) || data.length < 2 || data[0] != "questContentPort")
            return;

        data.shift();

        switch(data[0]) {
            case "connect":
                connect();
                break;
            case "disconnect":
                disconnect();
                port.disconnect();
                break;
            default:
                port.postMessage(data);
                break;
        }
    }

    function connect() {
        state = STATE_CONNECTED;
        decouplePort();
        port.onMessage.addListener(sendToExternal);
        port.onDisconnect.addListener(onPortDisconnect);
        port.postMessage([ CMD_CONTENT_SWITCH_TO_EXT ]);
    }

    function disconnect() {
        if(state == STATE_DISCONNECTED)
            throw new Error("Port has already disconnected!");

        window.removeEventListener(evtName, onExternalMessage);
    }

    function sendToExternal(data) {
        data.unshift("questExternalPort");
        window.postMessage(data, "*");
    }

    function onPortDisconnect() {
        disconnect();
        sendToExternal([ null ]);
    }

    function loadCheck() {
        if(isQuestInitiated())
            window.addEventListener(evtName, onExternalMessage);
        else
            requestAnimationFrame(loadCheck);
    }

    loadCheck();
}

function onStartup(isDev, isChrome, gamePath, folder, jsFolder, group, pageStyles) {
    if(isDev)
        alert("Now is the time to open the debug menu.");

    connected = true;

    const info = document.createElement("script"),
          loader = document.createElement("script");

    info.type = "questExtensionData";
    info.innerText = JSON.stringify({
        id: chrome.runtime.id,
        group: group,
        questPath: folder,
        gamePath: gamePath,
        styles: pageStyles,
        debug: isDev
    });
    loader.innerText = `import("${ jsFolder }quests/${ group }.js").then(module => new module.default()).catch(err => { alert("The quest could not be loaded.");${ isDev ? "" : " close();" } throw err; });`;
    document.head.appendChild(info);
    document.head.appendChild(loader);

    if(!isChrome)
        createEventPort();
}

function disconnect() {
    if(connected || isQuestInitiated())
        return;

    alert("Quests cannot be launched at this time.");
    close();
    location.href = "/";
}

main();